# This is used in pytest style locustfiles, see examples/test_pytest.py
from typing import TYPE_CHECKING

import pytest

if TYPE_CHECKING:
    from locust.user.users import User


class NoOpEvent:  # Fake locust.event.EventHook
    def fire(self, *, reverse=False, **kwargs):
        pass


_config: pytest.Config


# capture Config object instead of having to pass it explicitly to fixtures, which would make them more complex to call from Locust
def pytest_configure(config):
    global _config
    _config = config


@pytest.fixture
def session(user: "User | None" = None):
    # lazy import to avoid gevent monkey patching unless you actually use this fixture
    from locust.clients import HttpSession

    s = HttpSession(
        base_url=user.host if user else _config.getoption("--host"),
        request_event=user.environment.events.request if user else NoOpEvent(),
        user=user,
    )
    yield s
    s.close()


@pytest.fixture
def fastsession(user: "User | None" = None):
    # lazy import to avoid gevent monkey patching unless you actually use this fixture
    from locust.contrib.fasthttp import FastHttpSession

    s = FastHttpSession(
        base_url=user.host if user else _config.getoption("--host"),
        request_event=user.environment.events.request if user else NoOpEvent(),
        user=user,
    )
    yield s


def pytest_addoption(parser: pytest.Parser) -> None:
    parser.addoption("--host", "-H", action="store", default=None)



from locust import User

import gevent
import socketio


class SocketIOUser(User):
    """
    SocketIOUser wraps an instance of :class:`socketio.Client` to log requests.
    See example in :gh:`examples/socketio/socketio_ex.py`.
    """

    abstract = True
    options: dict = {}
    """socketio.Client options, e.g. `{"reconnection_attempts": 1, "reconnection_delay": 2}`"""
    client: socketio.Client
    """The underlying :class:`socketio.Client` instance. Can be useful to call directly if you want to skip logging a requests."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.client = socketio.Client(**self.options)
        self.ws_greenlet = gevent.spawn(self.client.wait)
        self.client.on("*", self.on_message)

    #
    def on_message(self, event: str, data: str) -> None:
        """
        This is the default handler for events. You can override it for custom behavior,
        or even register separate handlers using self.client.on(event, handler)

        Measuring response_time isn't obvious for for WebSockets. Sometimes a response time
        can be inferred from the event data (if it contains a timestamp) or related to
        a message that you sent. Override this method in your User class to do that.
        """
        self.environment.events.request.fire(
            request_type="WSR",
            name=event,
            response_time=0,
            response_length=len(data or []),
            exception=None,
            context={},
        )

    def connect(self, *args, **kwargs):
        """
        Wraps :meth:`socketio.Client.connect`.
        """
        with self.environment.events.request.measure("WS", "connect") as _:
            self.client.connect(*args, **kwargs)

    def send(self, name, data=None, namespace=None) -> None:
        """
        Wraps :meth:`socketio.Client.send`.
        """
        exception = None
        try:
            self.client.send(data, namespace)
        except Exception as e:
            exception = e
        self.environment.events.request.fire(
            request_type="WSS",
            name=name,
            response_time=0,
            response_length=len(data or []),
            exception=exception,
            context={},
        )

    def emit(self, name, data=None, namespace=None, callback=None) -> None:
        """
        Wraps :meth:`socketio.Client.emit`.
        """
        exception = None
        try:
            self.client.emit(name, data, namespace, callback)
        except Exception as e:
            exception = e
        self.environment.events.request.fire(
            request_type="WSE",
            name=name,
            response_time=0,
            response_length=len(data or []),
            exception=exception,
            context={},
        )

    def call(self, event, data=None, *args, **kwargs):
        """
        Wraps :meth:`socketio.Client.call`.
        """
        with self.environment.events.request.measure("WSC", event) as _:
            return self.client.call(event, data, *args, **kwargs)

    def on_stop(self):
        self.client.disconnect()



from __future__ import annotations

import re
import time
from contextlib import contextmanager
from typing import TYPE_CHECKING
from urllib.parse import urlparse, urlunparse

import requests
from requests import Response
from requests.adapters import HTTPAdapter
from requests.auth import HTTPBasicAuth
from requests.exceptions import InvalidSchema, InvalidURL, MissingSchema, RequestException
from urllib3 import PoolManager

from .exception import CatchResponseError, LocustError, ResponseError

if TYPE_CHECKING:
    import sys
    from collections.abc import Callable, Generator, Iterable, Mapping, MutableMapping
    from typing import Any, TypedDict

    from requests.cookies import RequestsCookieJar

    if sys.version_info >= (3, 11):
        from typing import Unpack
    else:
        from typing_extensions import Unpack

    # Annotations below were generated using output from mypy.
    # Mypy underneath uses information from the https://github.com/python/typeshed repo.

    class RequestKwargs(TypedDict, total=False):
        params: Any  # simplified signature
        headers: Mapping[str, str | bytes | None] | None
        cookies: RequestsCookieJar | MutableMapping[str, str] | None
        files: Any  # simplified signature
        auth: Any  # simplified signature
        timeout: float | tuple[float, float] | tuple[float, None] | None
        allow_redirects: bool
        proxies: MutableMapping[str, str] | None
        hooks: Mapping[str, Iterable[Callable[[Response], Any]] | Callable[[Response], Any]] | None
        stream: bool | None
        verify: bool | str | None
        cert: str | tuple[str, str] | None

    class RESTKwargs(RequestKwargs, total=False):
        name: str | None
        catch_response: bool
        context: dict


absolute_http_url_regexp = re.compile(r"^https?://", re.I)


class LocustResponse(Response):
    error: Exception | None = None

    def raise_for_status(self) -> None:
        if self.error:
            raise self.error
        Response.raise_for_status(self)


class HttpSession(requests.Session):
    """
    Class for performing web requests and holding (session-) cookies between requests (in order
    to be able to log in and out of websites). Each request is logged so that locust can display
    statistics.

    This is a slightly extended version of `python-request <https://requests.readthedocs.io/>`_'s
    :py:class:`requests.Session` class and mostly this class works exactly the same. However
    the methods for making requests (get, post, delete, put, head, options, patch, request)
    can now take a *url* argument that's only the path part of the URL, in which case the host
    part of the URL will be prepended with the HttpSession.base_url which is normally inherited
    from a User class' host attribute.

    Each of the methods for making requests also takes two additional optional arguments which
    are Locust specific and doesn't exist in python-requests. These are:

    :param name: (optional) An argument that can be specified to use as label in Locust's statistics instead of the URL path.
                 This can be used to group different URL's that are requested into a single entry in Locust's statistics.
    :param catch_response: (optional) Boolean argument that, if set, can be used to make a request return a context manager
                           to work as argument to a with statement. This will allow the request to be marked as a fail based on the content of the
                           response, even if the response code is ok (2xx). The opposite also works, one can use catch_response to catch a request
                           and then mark it as successful even if the response code was not (i.e 500 or 404).
    """

    def __init__(self, base_url, request_event, user, *args, pool_manager: PoolManager | None = None, **kwargs):
        super().__init__(*args, **kwargs)

        self.base_url = base_url
        self.request_event = request_event
        self.user = user

        # User can group name, or use the group context manager to gather performance statistics under a specific name
        # This is an alternative to passing in the "name" parameter to the requests function
        self.request_name: str | None = None

        # Check for basic authentication
        parsed_url = urlparse(self.base_url)
        if parsed_url.username and parsed_url.password:
            netloc = parsed_url.hostname
            if parsed_url.port:
                netloc += ":%d" % parsed_url.port

            # remove username and password from the base_url
            self.base_url = urlunparse(
                (parsed_url.scheme, netloc, parsed_url.path, parsed_url.params, parsed_url.query, parsed_url.fragment)
            )
            # configure requests to use basic auth
            self.auth = HTTPBasicAuth(parsed_url.username, parsed_url.password)

        self.mount("https://", LocustHttpAdapter(pool_manager=pool_manager))
        self.mount("http://", LocustHttpAdapter(pool_manager=pool_manager))

    def _build_url(self, path) -> str:
        """prepend url with hostname unless it's already an absolute URL"""
        if absolute_http_url_regexp.match(path):
            return path
        else:
            return f"{self.base_url}{path}"

    @contextmanager
    def rename_request(self, name: str) -> Generator[None]:
        """Group requests using the "with" keyword"""

        self.request_name = name
        try:
            yield
        finally:
            self.request_name = None

    def request(  # type: ignore[override]
        self,
        method: str | bytes,
        url: str | bytes,
        name: str | None = None,
        catch_response: bool = False,
        context: dict = {},
        *,
        data: Any = None,
        json: Any = None,
        **kwargs: Unpack[RequestKwargs],
    ):
        """
        Constructs and sends a :py:class:`requests.Request`.
        Returns :py:class:`requests.Response` object.

        :param method: method for the new :class:`Request` object.
        :param url: URL for the new :class:`Request` object.
        :param name: (optional) An argument that can be specified to use as label in Locust's statistics instead of the URL path.
          This can be used to group different URL's that are requested into a single entry in Locust's statistics.
        :param catch_response: (optional) Boolean argument that, if set, can be used to make a request return a context manager
          to work as argument to a with statement. This will allow the request to be marked as a fail based on the content of the
          response, even if the response code is ok (2xx). The opposite also works, one can use catch_response to catch a request
          and then mark it as successful even if the response code was not (i.e 500 or 404).
        :param params: (optional) Dictionary or bytes to be sent in the query string for the :class:`Request`.
        :param data: (optional) Dictionary, list of tuples, bytes, or file-like object to send in the body of the :class:`Request`.
        :param json: (optional) json to send in the body of the :class:`Request`.
        :param headers: (optional) Dictionary of HTTP Headers to send with the :class:`Request`.
        :param cookies: (optional) Dict or CookieJar object to send with the :class:`Request`.
        :param files: (optional) Dictionary of ``'filename': file-like-objects`` for multipart encoding upload.
        :param auth: (optional) Auth tuple or callable to enable Basic/Digest/Custom HTTP Auth.
        :param timeout: (optional) How long to wait for the server to send data before giving up, as a float, or a :ref:`(connect timeout, read timeout) <timeouts>` tuple.
        :type timeout: float or tuple
        :param allow_redirects: (optional) Set to True by default.
        :type allow_redirects: bool
        :param proxies: (optional) Dictionary mapping protocol or protocol and hostname to the URL of the proxy.
        :param hooks: (optional) Dictionary mapping hook name to one event or list of events, event must be callable.
        :param stream: (optional) whether to immediately download the response content. Defaults to ``False``.
        :param verify: (optional) Either a boolean, in which case it controls whether we verify
          the server's TLS certificate, or a string, in which case it must be a path
          to a CA bundle to use. Defaults to ``True``. When set to
          ``False``, requests will accept any TLS certificate presented by
          the server, and will ignore hostname mismatches and/or expired
          certificates, which will make your application vulnerable to
          man-in-the-middle (MitM) attacks. Setting verify to ``False``
          may be useful during local development or testing.
        :param cert: (optional) if String, path to ssl client cert file (.pem). If Tuple, ('cert', 'key') pair.
        """

        # if group name has been set and no name parameter has been passed in; set the name parameter to group_name
        if self.request_name and not name:
            name = self.request_name

        # prepend url with hostname unless it's already an absolute URL
        complete_url = self._build_url(url)

        start_time = time.time()
        start_perf_counter = time.perf_counter()
        response = self._send_request_safe_mode(method, complete_url, data=data, json=json, **kwargs)
        response_time = (time.perf_counter() - start_perf_counter) * 1000

        if request_before_redirect := (response.history and response.history[0] or response).request:
            complete_url = str(request_before_redirect.url)
            if not name:
                name = request_before_redirect.path_url
        else:  # in rare cases, such as ProtocolError('Response ended prematurely'), response.request may be None
            if not name:
                name = str(url)

        if self.user:
            context = {**self.user.context(), **context}

        # store meta data that is used when reporting the request to locust's statistics
        request_meta = {
            "request_type": method,
            "response_time": response_time,
            "name": name,
            "context": context,
            "response": response,
            "exception": None,
            "start_time": start_time,
            "url": complete_url,
        }

        # get the length of the content, but if the argument stream is set to True, we take
        # the size from the content-length header, in order to not trigger fetching of the body
        if kwargs.get("stream", False):
            request_meta["response_length"] = int(response.headers.get("content-length") or 0)
        else:
            request_meta["response_length"] = len(response.content or b"")

        if catch_response:
            return ResponseContextManager(response, request_event=self.request_event, request_meta=request_meta)
        else:
            with ResponseContextManager(response, request_event=self.request_event, request_meta=request_meta):
                pass
            return response

    def _send_request_safe_mode(self, method, url, **kwargs) -> Response | LocustResponse:
        """
        Send an HTTP request, and catch any exception that might occur due to connection problems.

        Safe mode has been removed from requests 1.x.
        """
        try:
            return super().request(method, url, **kwargs)
        except (MissingSchema, InvalidSchema, InvalidURL):
            raise
        except RequestException as e:
            r = LocustResponse()
            r.error = e
            r.status_code = 0
            r.request = e.request  # type: ignore
            return r

    def get(
        self, url: str | bytes, *, data: Any = None, json: Any = None, **kwargs: Unpack[RESTKwargs]
    ) -> ResponseContextManager | Response | LocustResponse:
        """Sends a GET request"""
        kwargs.setdefault("allow_redirects", True)
        return self.request("GET", url, data=data, json=json, **kwargs)

    def options(
        self,
        url: str | bytes,
        *,
        data: Any = None,
        json: Any = None,
        **kwargs: Unpack[RESTKwargs],
    ) -> ResponseContextManager | Response | LocustResponse:
        """Sends a OPTIONS request"""
        kwargs.setdefault("allow_redirects", True)
        return self.request("OPTIONS", url, data=data, json=json, **kwargs)

    def head(
        self,
        url: str | bytes,
        *,
        data: Any = None,
        json: Any = None,
        **kwargs: Unpack[RESTKwargs],
    ) -> ResponseContextManager | Response | LocustResponse:
        """Sends a HEAD request"""
        kwargs.setdefault("allow_redirects", False)
        return self.request("HEAD", url, data=data, json=json, **kwargs)

    def post(
        self,
        url: str | bytes,
        data: Any = None,
        json: Any = None,
        **kwargs: Unpack[RESTKwargs],
    ) -> ResponseContextManager | Response | LocustResponse:
        """Sends a POST request"""
        return self.request("POST", url, data=data, json=json, **kwargs)

    def put(
        self,
        url: str | bytes,
        data: Any = None,
        *,
        json: Any = None,
        **kwargs: Unpack[RESTKwargs],
    ) -> ResponseContextManager | Response | LocustResponse:
        """Sends a PUT request"""
        return self.request("PUT", url, data=data, json=json, **kwargs)

    def patch(
        self,
        url: str | bytes,
        data: Any = None,
        *,
        json: Any = None,
        **kwargs: Unpack[RESTKwargs],
    ) -> ResponseContextManager | Response | LocustResponse:
        """Sends a PATCH request"""
        return self.request("PATCH", url, data=data, json=json, **kwargs)

    def delete(
        self,
        url: str | bytes,
        *,
        data: Any = None,
        json: Any = None,
        **kwargs: Unpack[RESTKwargs],
    ) -> ResponseContextManager | Response | LocustResponse:
        """Sends a DELETE request"""
        return self.request("DELETE", url, data=data, json=json, **kwargs)


class ResponseContextManager(LocustResponse):
    """
    A Response class that also acts as a context manager that provides the ability to manually
    control if an HTTP request should be marked as successful or a failure in Locust's statistics

    This class is a subclass of :py:class:`Response <requests.Response>` with two additional
    methods: :py:meth:`success <locust.clients.ResponseContextManager.success>` and
    :py:meth:`failure <locust.clients.ResponseContextManager.failure>`.
    """

    _manual_result: bool | Exception | None = None
    _entered = False

    def __init__(self, response, request_event, request_meta):
        # copy data from response to this object
        self.__dict__ = response.__dict__
        self._request_event = request_event
        self.request_meta = request_meta

    def __enter__(self):
        self._entered = True
        return self

    def __exit__(self, exc, value, traceback):  # type: ignore[override]
        # if the user has already manually marked this response as failure or success
        # we can ignore the default behaviour of letting the response code determine the outcome
        if self._manual_result is not None:
            if self._manual_result is True:
                self.request_meta["exception"] = None
            elif isinstance(self._manual_result, Exception):
                self.request_meta["exception"] = self._manual_result
            self._report_request()
            return exc is None

        if exc:
            if isinstance(value, ResponseError):
                self.request_meta["exception"] = value
                self._report_request()
            else:
                # we want other unknown exceptions to be raised
                return False
        else:
            # Since we use the Exception message when grouping failures, in order to not get
            # multiple failure entries for different URLs for the same name argument, we need
            # to temporarily override the response.url attribute
            orig_url = self.url
            self.url = self.request_meta["name"]

            try:
                self.raise_for_status()
            except requests.exceptions.RequestException as e:
                while (
                    isinstance(
                        e,
                        requests.exceptions.ConnectionError
                        | requests.packages.urllib3.exceptions.ProtocolError
                        | requests.packages.urllib3.exceptions.MaxRetryError
                        | requests.packages.urllib3.exceptions.NewConnectionError,
                    )
                    and e.__context__  # Not sure if the above exceptions can ever be the lowest level, but it is good to be sure
                ):
                    e = e.__context__
                self.request_meta["exception"] = e

            self._report_request()
            self.url = orig_url

        return True

    def _report_request(self, exc=None):
        self._request_event.fire(**self.request_meta)

    def success(self):
        """
        Report the response as successful

        Example::

            with self.client.get("/does/not/exist", catch_response=True) as response:
                if response.status_code == 404:
                    response.success()
        """
        if not self._entered:
            raise LocustError(
                "Tried to set status on a request that has not yet been made. Make sure you use a with-block, like this:\n\nwith self.client.request(..., catch_response=True) as response:\n    response.success()"
            )
        self._manual_result = True

    def failure(self, exc):
        """
        Report the response as a failure.

        if exc is anything other than a python exception (like a string) it will
        be wrapped inside a CatchResponseError.

        Example::

            with self.client.get("/", catch_response=True) as response:
                if response.content == b"":
                    response.failure("No data")
        """
        if not self._entered:
            raise LocustError(
                "Tried to set status on a request that has not yet been made. Make sure you use a with-block, like this:\n\nwith self.client.request(..., catch_response=True) as response:\n    response.failure(...)"
            )
        if not isinstance(exc, Exception):
            exc = CatchResponseError(exc)
        self._manual_result = exc


class LocustHttpAdapter(HTTPAdapter):
    def __init__(self, pool_manager: PoolManager | None, *args, **kwargs):
        self.poolmanager = pool_manager
        super().__init__(*args, **kwargs)

    def init_poolmanager(self, *args, **kwargs):
        if self.poolmanager is None:
            super().init_poolmanager(*args, **kwargs)


# Monkey patch Response class to give some guidance
def _missing_catch_response_True(self, *_args, **_kwargs):
    raise LocustError(
        "If you want to change the state of the request using .success() or .failure(), you must pass catch_response=True. See http://docs.locust.io/en/stable/writing-a-locustfile.html#validating-responses"
    )


Response.success = _missing_catch_response_True  # type: ignore[attr-defined]
Response.failure = _missing_catch_response_True  # type: ignore[attr-defined]



import os

if os.getenv("LOCUST_PLAYWRIGHT", None):
    print("LOCUST_PLAYWRIGHT setting is no longer needed (because locust-plugins no longer installs trio)")
    print("Uninstall trio package and remove the setting.")
    try:
        # preserve backwards compatibility for now
        import trio  # noqa: F401
    except ModuleNotFoundError:
        # dont show a massive callstack if trio is not installed
        os._exit(1)

if not os.getenv("LOCUST_SKIP_MONKEY_PATCH", None):
    from gevent import monkey, queue

    monkey.patch_all()

    if not os.getenv("LOCUST_SKIP_URLLIB3_PATCH", None):
        import urllib3

        urllib3.connectionpool.ConnectionPool.QueueCls = queue.LifoQueue
        # https://github.com/locustio/locust/issues/2812

from ._version import version as __version__
from .contrib.fasthttp import FastHttpUser
from .debug import run_single_user
from .event import Events
from .shape import LoadTestShape
from .user import wait_time
from .user.markov_taskset import MarkovTaskSet, transition, transitions
from .user.sequential_taskset import SequentialTaskSet
from .user.task import TaskSet, tag, task
from .user.users import HttpUser, User
from .user.wait_time import between, constant, constant_pacing, constant_throughput

events = Events()

__all__ = (
    "SequentialTaskSet",
    "MarkovTaskSet",
    "transition",
    "transitions",
    "wait_time",
    "task",
    "tag",
    "TaskSet",
    "HttpUser",
    "FastHttpUser",
    "User",
    "between",
    "constant",
    "constant_pacing",
    "constant_throughput",
    "events",
    "LoadTestShape",
    "run_single_user",
    "HttpLocust",
    "Locust",
    "__version__",
)

# Used for raising a DeprecationWarning if old Locust/HttpLocust is used
from .util.deprecation import DeprecatedHttpLocustClass as HttpLocust
from .util.deprecation import DeprecatedLocustClass as Locust



from locust.exception import LocustError
from locust.user.task import TaskSetMeta
from locust.user.users import TaskSet

import logging
import random
from collections.abc import Callable

MarkovTaskT = Callable[..., None]


class NoMarkovTasksError(LocustError):
    """Raised when a MarkovTaskSet class doesn't define any Markov tasks."""

    pass


class InvalidTransitionError(LocustError):
    """Raised when a transition in a MarkovTaskSet points to a non-existent task."""

    pass


class NonMarkovTaskTransitionError(LocustError):
    """Raised when a transition in a MarkovTaskSet points to a task that doesn't define transitions."""

    pass


class MarkovTaskTagError(LocustError):
    """Raised when tags are used with Markov tasks, which is unsupported."""

    pass


def is_markov_task(task: MarkovTaskT):
    """
    Determines if a task is a Markov task by checking if it has transitions defined.

    :param task: The task to check
    :return: True if the task is a Markov task, False otherwise
    """
    return "transitions" in dir(task)


def transition(func_name: str, weight: int = 1) -> Callable[[MarkovTaskT], MarkovTaskT]:
    """
    Decorator for adding a single transition to a Markov task.

    This decorator allows you to define a transition from one task to another in a MarkovTaskSet,
    with an associated weight that determines the probability of taking this transition.

    :param func_name: The name of the target task function
    :param weight: The weight of this transition (default: 1)
    :return: The decorated function with the transition added

    Example::

        class UserBehavior(MarkovTaskSet):
            @transition('browse_products')
            def index(self):
                self.client.get("/")

            @transition('index', weight=3)
            @transition('product_page', weight=1)
            def browse_products(self):
                self.client.get("/products/")
    """

    def decorator_func(decorated):
        if not hasattr(decorated, "transitions"):
            decorated.transitions = {}

        decorated.transitions[func_name] = weight
        return decorated

    return decorator_func


def transitions(weights: dict[str, int] | list[tuple[str, int] | str]) -> Callable[[MarkovTaskT], MarkovTaskT]:
    """
    Decorator for adding multiple transitions to a Markov task at once.

    This decorator allows you to define multiple transitions from one task to others in a MarkovTaskSet,
    with associated weights that determine the probability of taking each transition.

    :param weights: Either a dictionary mapping function names to weights, or a list of function names
                   (with default weight 1) or (function_name, weight) tuples
    :return: The decorated function with the transitions added

    Example::

        class UserBehavior(MarkovTaskSet):
            @transitions({'checkout': 1, 'browse_products': 3, 'index': 2})
            def view_cart(self):
                self.client.get("/cart/")

            @transitions([
                ('index', 2),      # with weight 2
                'browse_products'  # with default weight 1
            ])
            def checkout(self):
                self.client.get("/checkout/")
    """

    def parse_list_item(item: tuple[str, int] | str) -> tuple[str, int]:
        return item if isinstance(item, tuple) else (item, 1)

    def decorator_func(decorated):
        if not hasattr(decorated, "transitions"):
            decorated.transitions = {}

        decorated.transitions.update(
            weights
            if isinstance(weights, dict)
            else {func_name: weight for func_name, weight in map(parse_list_item, weights)}
        )

        return decorated

    return decorator_func


def get_markov_tasks(class_dict: dict) -> list:
    """
    Extracts all Markov tasks from a class dictionary.

    This function is used internally by MarkovTaskSetMeta to find all methods
    that have been decorated with @transition or @transitions.

    :param class_dict: Dictionary containing class attributes and methods
    :return: List of functions that are Markov tasks
    """
    return [fn for fn in class_dict.values() if is_markov_task(fn)]


def to_weighted_list(transitions: dict):
    return [name for name in transitions.keys() for _ in range(transitions[name])]


def validate_has_markov_tasks(tasks: list, classname: str):
    """
    Validates that a MarkovTaskSet has at least one Markov task.

    This function is used internally during MarkovTaskSet validation to ensure
    that the class has at least one method decorated with @transition or @transitions.

    :param tasks: List of tasks to validate
    :param classname: Name of the class being validated (for error messages)
    :raises NoMarkovTasksError: If no Markov tasks are found
    """
    if not tasks:
        raise NoMarkovTasksError(
            f"No Markov tasks defined in class {classname}. Use the @transition(s) decorators to define some."
        )


def validate_transitions(tasks: list, class_dict: dict, classname: str):
    """
    Validates that all transitions in Markov tasks point to existing Markov tasks.

    This function checks two conditions for each transition:
    1. The target task exists in the class
    2. The target task is also a Markov task (has transitions defined)

    :param tasks: List of Markov tasks to validate
    :param class_dict: Dictionary containing class attributes and methods
    :param classname: Name of the class being validated (for error messages)
    :raises InvalidTransitionError: If a transition points to a non-existent task
    :raises NonMarkovTaskTransitionError: If a transition points to a task that isn't a Markov task
    """
    for task in tasks:
        for dest in task.transitions.keys():
            dest_task = class_dict.get(dest)
            if not dest_task:
                raise InvalidTransitionError(
                    f"Transition to {dest} from {task.__name__} is invalid since no such element exists on class {classname}"
                )
            if not is_markov_task(dest_task):
                raise NonMarkovTaskTransitionError(
                    f"{classname}.{dest} cannot be used as a target for a transition since it does not define any transitions of its own."
                    + f"Used as a transition from {task.__name__}."
                )


def validate_no_unreachable_tasks(tasks: list, class_dict: dict, classname: str):
    """
    Checks for and warns about unreachable Markov tasks in a MarkovTaskSet.

    This function uses depth-first search (DFS) starting from the first task to identify
    all reachable tasks. It then warns about any tasks that cannot be reached from the
    starting task through the defined transitions.

    :param tasks: List of Markov tasks to validate
    :param class_dict: Dictionary containing class attributes and methods
    :param classname: Name of the class being validated (for warning messages)
    :return: The original list of tasks
    """
    visited = set()

    def dfs(task_name):
        visited.add(task_name)
        # Convert to a weighted list first to handle bad weights
        for dest in set(to_weighted_list(class_dict.get(task_name).transitions)):
            if dest not in visited:
                dfs(dest)

    dfs(tasks[0].__name__)
    unreachable = set([task.__name__ for task in tasks]) - visited

    if len(unreachable) > 0:
        logging.warning(f"The following markov tasks are unreachable in class {classname}: {unreachable}")

    return tasks


def validate_no_tags(task, classname: str):
    """
    Validates that Markov tasks don't have tags, which are unsupported.

    Tags are not supported for MarkovTaskSet because they can make the Markov chain invalid
    by potentially filtering out tasks that are part of the chain.

    :param task: The task to validate
    :param classname: Name of the class being validated (for error messages)
    :raises MarkovTaskTagError: If the task has tags
    """
    if "locust_tag_set" in dir(task):
        raise MarkovTaskTagError(
            "Tags are unsupported for MarkovTaskSet since they can make the markov chain invalid. "
            + f"Tags detected on {classname}.{task.__name__}: {task.locust_tag_set}"
        )


def validate_task_name(decorated_func):
    """
    Validates that certain method names aren't used as Markov tasks.

    This function checks for special method names that shouldn't be used as Markov tasks:
    - "on_stop" and "on_start": Using these as Markov tasks will cause them to be called
      both as tasks AND on stop/start, which is usually not what the user intended.
    - "run": This method is used internally by Locust and must not be overridden or
      annotated with transitions.

    :param decorated_func: The function to validate
    :raises Exception: If the function name is "run"
    """
    if decorated_func.__name__ in ["on_stop", "on_start"]:
        logging.warning(
            "You have tagged your on_stop/start function with @transition. This will make the method get called both as a step AND on stop/start."
        )  # this is usually not what the user intended
    if decorated_func.__name__ == "run":
        raise Exception(
            "TaskSet.run() is a method used internally by Locust, and you must not override it or annotate it with transitions"
        )


def validate_markov_chain(tasks: list, class_dict: dict, classname: str):
    """
    Runs all validation functions on a Markov chain.

    :param tasks: List of Markov tasks to validate
    :param class_dict: Dictionary containing class attributes and methods
    :param classname: Name of the class being validated (for error/warning messages)
    :raises: Various exceptions if validation fails
    """
    validate_has_markov_tasks(tasks, classname)
    validate_transitions(tasks, class_dict, classname)
    validate_no_unreachable_tasks(tasks, class_dict, classname)
    for task in tasks:
        validate_task_name(task)
        validate_no_tags(task, classname)


class MarkovTaskSetMeta(TaskSetMeta):
    """
    Meta class for MarkovTaskSet. It's used to allow MarkovTaskSet classes to specify
    task execution using the @transition(s) decorators
    """

    def __new__(mcs, classname, bases, class_dict):
        if not class_dict.get("abstract"):
            class_dict["abstract"] = False

            tasks = get_markov_tasks(class_dict)
            validate_markov_chain(tasks, class_dict, classname)
            class_dict["current"] = tasks[0]
            for task in tasks:
                task.transitions = to_weighted_list(task.transitions)

        return type.__new__(mcs, classname, bases, class_dict)


class MarkovTaskSet(TaskSet, metaclass=MarkovTaskSetMeta):
    """
    Class defining a probabilistic sequence of functions that a User will execute.
    The sequence is defined by a Markov Chain to describe a user's load.
    It holds a current state and a set of possible transitions for each state.
    Every transition as an associated weight that defines how likely it is to be taken.
    """

    current: Callable | TaskSet

    abstract: bool = True
    """If abstract is True, the class is meant to be subclassed, and the markov chain won't be validated"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def get_next_task(self):
        """
        Gets the next task to execute based on the current state and transitions.

        :return: The current task to execute
        """
        fn = self.current

        transitions = getattr(fn, "transitions")
        next = random.choice(transitions)
        self.current = getattr(self, next)

        return fn



import re
from datetime import timedelta


def parse_timespan(time_str) -> int:
    """
    Parse a string representing a time span and return the number of seconds.
    Valid formats are: 20, 20s, 3m, 2h, 1h20m, 3h30m10s, etc.
    """
    if not time_str:
        raise ValueError("Invalid time span format")

    if re.match(r"^\d+$", time_str):
        # if an int is specified we assume they want seconds
        return int(time_str)

    timespan_regex = re.compile(r"((?P<hours>\d+?)h)?((?P<minutes>\d+?)m)?((?P<seconds>\d+?)s)?")
    parts = timespan_regex.match(time_str)
    if not parts:
        raise ValueError("Invalid time span format. Valid formats: 20, 20s, 3m, 2h, 1h20m, 3h30m10s, etc.")
    time_params = {name: int(value) for name, value in parts.groupdict().items() if value}
    if not time_params:
        raise ValueError("Invalid time span format. Valid formats: 20, 20s, 3m, 2h, 1h20m, 3h30m10s, etc.")
    return int(timedelta(**time_params).total_seconds())



from locust import User, events

import time

from pymongo import MongoClient
from pymongo.errors import PyMongoError


class MongoDBClient(MongoClient):
    def __init__(self, conn_string, db_name):
        super().__init__(conn_string)
        self.db = self.client[db_name]

    def execute_query(self, collection_name, query):
        start_time = time.time()
        try:
            collection = self.db[collection_name]
            collection.find(query)

            response_time = int((time.time() - start_time) * 1000)
            events.request.fire(request_type="MONGODB", name="QUERY", response_time=response_time, response_length=0)
        except PyMongoError as e:
            response_time = int((time.time() - start_time) * 1000)
            events.request.fire(
                request_type="MONGODB",
                name="QUERY",
                response_time=response_time,
                response_length=0,
                exception=e,
            )


class MongoDBUser(User):
    abstract = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.client = MongoDBClient(conn_string=self.conn_string, db_name=self.db_name)

    def on_stop(self):
        self.client.close()



from __future__ import annotations

import importlib
import importlib.util
import inspect
import os
import sys

import pytest
from _pytest.config import Config

from ..shape import LoadTestShape
from ..user import User
from ..user.users import PytestUser


def is_user_class(item) -> bool:
    """
    Check if a variable is a runnable (non-abstract) User class
    """
    return bool(inspect.isclass(item) and issubclass(item, User) and item.abstract is False)


def is_shape_class(item) -> bool:
    """
    Check if a class is a LoadTestShape
    """
    return bool(inspect.isclass(item) and issubclass(item, LoadTestShape) and not getattr(item, "abstract", True))


def load_locustfile(path) -> tuple[dict[str, type[User]], list[LoadTestShape]]:
    """
    Import given locustfile path and return (docstring, callables).

    Specifically, the locustfile's ``__doc__`` attribute (a string) and a
    dictionary of ``{'name': callable}`` containing all callables which pass
    the "is a Locust" test.
    """

    # Start with making sure the current working dir is in the sys.path
    sys.path.insert(0, os.getcwd())
    # Get directory and locustfile name
    directory, locustfile = os.path.split(path)
    # If the directory isn't in the PYTHONPATH, add it so our import will work
    added_to_path = False
    index = None
    if directory not in sys.path:
        sys.path.insert(0, directory)
        added_to_path = True
    # If the directory IS in the PYTHONPATH, move it to the front temporarily,
    # otherwise other locustfiles -- like Locusts's own -- may scoop the intended
    # one.
    else:
        i = sys.path.index(directory)
        if i != 0:
            # Store index for later restoration
            index = i
            # Add to front, then remove from original position
            sys.path.insert(0, directory)
            del sys.path[i + 1]

    # Perform the import
    module_name = os.path.splitext(locustfile)[0]
    if module_name == "locust":
        module_name = "locustfile"  # Avoid conflict with locust package
    loader = importlib.machinery.SourceFileLoader(module_name, path)
    spec = importlib.util.spec_from_file_location(module_name, path, loader=loader)
    if spec is None:
        sys.stderr.write(f"Unable to get module spec for {module_name} in {path}")
        sys.exit(1)

    imported = importlib.util.module_from_spec(spec)
    sys.modules[imported.__name__] = imported
    loader.exec_module(imported)

    # Remove directory from path if we added it ourselves (just to be neat)
    if added_to_path:
        del sys.path[0]
    # Put back in original index if we moved it
    if index is not None:
        sys.path.insert(index + 1, directory)
        del sys.path[0]
    # Return our two-tuple
    user_classes = {name: value for name, value in vars(imported).items() if is_user_class(value)}

    # Find shape class, if any, return it
    shape_classes = [value() for value in vars(imported).values() if is_shape_class(value)]
    return user_classes, shape_classes


def load_locustfile_pytest(path) -> dict[str, type[User]]:
    """
    Create User classes from pytest test functions.

    It relies on some pytest internals to collect test functions and their fixtures,
    but it should be reasonably stable for simple use cases.

    Fixtures (like `session` and `fastsession`) are defined in `pytestplugin.py`

    See `examples/test_pytest.py` and `locust/test/test_pytest_locustfile.py`
    """
    user_classes: dict[str, type[PytestUser]] = {}
    # collect tests and set up fixture manager
    config = Config.fromdictargs({}, [path])
    config._do_configure()
    session = pytest.Session.from_config(config)
    config.hook.pytest_sessionstart(session=session)
    session.perform_collect()
    config.hook.pytest_collection_modifyitems(session=session, config=config, items=session.items)
    fm = session._fixturemanager

    for function in session.items:
        if isinstance(function, pytest.Function):
            sig = inspect.signature(function.obj)
            function.kwargs = {}  # type: ignore[attr-defined]
            for name in sig.parameters:
                defs = fm.getfixturedefs(name, function)
                if not defs:
                    raise ValueError(f"Could not find fixture for parameter {name!r} in {function.name}")
                if len(defs) > 1:
                    raise ValueError(f"Multiple fixtures found for parameter {name!r} in {function.name}: {defs}")
                function.fixturedef = defs[0]  # type: ignore[attr-defined]
            if not function.name in user_classes:
                user_classes[function.name] = type(function.name, (PytestUser,), {})
                user_classes[function.name].functions = []
            user_classes[function.name].functions.append(function)
        else:
            pass  # Skipping non-function item
    return user_classes  # type: ignore



# This directory contains modules that are experimental.

# Generally they should not have any significant issues, but they may change without a major version bump for locust.

# Some modules may require optional dependencies, typically installed with `pip install locust[name-of-module]`.



# Note: this User is experimental and may change without notice.
# The filename is oai.py so it doesnt clash with the openai package.
from locust.user import User

import os
import time
from collections.abc import Generator
from contextlib import contextmanager

import httpx
from openai import OpenAI  # dont forget to install openai

# convenience for when running in Locust Cloud, where only LOCUST_* env vars are forwarded
if "LOCUST_OPENAI_API_KEY" in os.environ:
    os.environ["OPENAI_API_KEY"] = os.environ["LOCUST_OPENAI_API_KEY"]

if not "OPENAI_API_KEY" in os.environ:
    raise Exception("You need to set OPENAI_API_KEY or LOCUST_OPENAI_API_KEY env var to use OpenAIUser")


class OpenAIClient(OpenAI):
    def __init__(self, request_event, user, *args, **kwargs):
        self.request_name = None  # used to override url-based request names
        self.user = user  # currently unused, but could be useful later

        def request_start(request):
            request.start_time = time.time()
            request.start_perf_counter = time.perf_counter()

        def request_end(response):
            exception = None
            response.read()
            response_time = (time.perf_counter() - response.request.start_perf_counter) * 1000
            try:
                response.raise_for_status()
            except httpx.HTTPStatusError as e:
                exception = e
            request_event.fire(
                request_type=response.request.method,
                name=self.request_name if self.request_name else response.url.path,
                context={},
                response=response,
                exception=exception,
                start_time=response.request.start_time,
                response_time=response_time,
                # Store the number of output tokens as response_length instead of the actual payload size because it is more useful
                response_length=response.json().get("usage", {}).get("output_tokens", 0),
                url=response.url,
            )

        super().__init__(
            *args,
            **kwargs,
            http_client=httpx.Client(event_hooks={"request": [request_start], "response": [request_end]}),
        )

    @contextmanager
    def rename_request(self, name: str) -> Generator[None]:
        """Group requests using the "with" keyword"""

        self.request_name = name
        try:
            yield
        finally:
            self.request_name = None


class OpenAIUser(User):
    abstract = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.client = OpenAIClient(self.environment.events.request, user=self)



import gevent.monkey

gevent.monkey.patch_all()
import grpc.experimental.gevent as grpc_gevent

grpc_gevent.init_gevent()

from locust import User, events

import time
from abc import ABC, abstractmethod
from typing import Any

from pymilvus import CollectionSchema, MilvusClient
from pymilvus.milvus_client import IndexParams


class BaseClient(ABC):
    @abstractmethod
    def close(self) -> None:
        pass

    @abstractmethod
    def create_collection(self, schema, index_params) -> None:
        pass

    @abstractmethod
    def insert(self, data) -> dict[str, Any]:
        pass

    @abstractmethod
    def upsert(self, data) -> dict[str, Any]:
        pass

    @abstractmethod
    def search(
        self,
        data,
        anns_field,
        limit,
        filter="",
        search_params=None,
        output_fields=None,
        calculate_recall=False,
        ground_truth=None,
    ) -> dict[str, Any]:
        pass

    @abstractmethod
    def hybrid_search(self, reqs, ranker, limit, output_fields=None) -> dict[str, Any]:
        pass

    @abstractmethod
    def query(self, filter, output_fields=None) -> dict[str, Any]:
        pass

    @abstractmethod
    def delete(self, filter) -> dict[str, Any]:
        pass


class MilvusV2Client(BaseClient):
    """Milvus v2 Python SDK Client Wrapper"""

    def __init__(self, uri, collection_name, token="root:Milvus", db_name="default", timeout=60):
        self.uri = uri
        self.collection_name = collection_name
        self.token = token
        self.db_name = db_name
        self.timeout = timeout

        # Initialize MilvusClient v2
        self.client = MilvusClient(
            uri=self.uri,
            token=self.token,
            db_name=self.db_name,
            timeout=self.timeout,
        )

    def close(self):
        self.client.close()

    def create_collection(self, schema, index_params):
        self.client.create_collection(
            collection_name=self.collection_name,
            schema=schema,
            index_params=index_params,
        )

    def insert(self, data):
        start = time.time()
        try:
            result = self.client.insert(collection_name=self.collection_name, data=data)
            total_time = (time.time() - start) * 1000
            return {"success": True, "response_time": total_time, "result": result}
        except Exception as e:
            return {
                "success": False,
                "response_time": (time.time() - start) * 1000,
                "exception": e,
            }

    def upsert(self, data):
        start = time.time()
        try:
            result = self.client.upsert(collection_name=self.collection_name, data=data)
            total_time = (time.time() - start) * 1000
            return {"success": True, "response_time": total_time, "result": result}
        except Exception as e:
            return {
                "success": False,
                "response_time": (time.time() - start) * 1000,
                "exception": e,
            }

    def search(
        self,
        data,
        anns_field,
        limit,
        filter="",
        search_params=None,
        output_fields=None,
        calculate_recall=False,
        ground_truth=None,
    ):
        if output_fields is None:
            output_fields = ["id"]

        start = time.time()
        try:
            result = self.client.search(
                collection_name=self.collection_name,
                data=data,
                anns_field=anns_field,
                filter=filter,
                limit=limit,
                search_params=search_params,
                output_fields=output_fields,
            )
            total_time = (time.time() - start) * 1000
            empty = len(result) == 0 or all(len(r) == 0 for r in result)

            # Prepare base result
            search_result = {
                "success": not empty,
                "response_time": total_time,
                "empty": empty,
                "result": result,
            }

            # Calculate recall if requested
            if calculate_recall and ground_truth is not None and not empty:
                recall_value = self.get_recall(result, ground_truth, limit)
                search_result["recall"] = recall_value

            return search_result
        except Exception as e:
            return {
                "success": False,
                "response_time": (time.time() - start) * 1000,
                "exception": e,
            }

    def hybrid_search(self, reqs, ranker, limit, output_fields=None):
        if output_fields is None:
            output_fields = ["id"]

        start = time.time()
        try:
            result = self.client.hybrid_search(
                collection_name=self.collection_name,
                reqs=reqs,
                ranker=ranker,
                limit=limit,
                output_fields=output_fields,
                timeout=self.timeout,
            )
            total_time = (time.time() - start) * 1000
            empty = len(result) == 0 or all(len(r) == 0 for r in result)

            # Prepare base result
            search_result = {
                "success": not empty,
                "response_time": total_time,
                "empty": empty,
                "result": result,
            }

            return search_result
        except Exception as e:
            return {
                "success": False,
                "response_time": (time.time() - start) * 1000,
                "exception": e,
            }

    @staticmethod
    def get_recall(search_results, ground_truth, limit=None):
        """Calculate recall for V2 client search results."""
        try:
            # Extract IDs from V2 search results
            retrieved_ids = []
            if isinstance(search_results, list) and len(search_results) > 0:
                # search_results[0] contains the search results for the first query
                for hit in search_results[0] if isinstance(search_results[0], list) else search_results:
                    if isinstance(hit, dict) and "id" in hit:
                        retrieved_ids.append(hit["id"])
                    elif hasattr(hit, "get"):
                        retrieved_ids.append(hit.get("id"))

            # Apply limit if specified
            if limit is None:
                limit = len(retrieved_ids)

            if len(ground_truth) < limit:
                raise ValueError(f"Ground truth length is less than limit: {len(ground_truth)} < {limit}")

            # Calculate recall
            ground_truth_set = set(ground_truth[:limit])
            retrieved_set = set(retrieved_ids)
            intersect = len(ground_truth_set.intersection(retrieved_set))
            return intersect / len(ground_truth_set)

        except Exception:
            return 0.0

    def query(self, filter, output_fields=None):
        if output_fields is None:
            output_fields = ["id"]

        start = time.time()
        try:
            result = self.client.query(
                collection_name=self.collection_name,
                filter=filter,
                output_fields=output_fields,
            )
            total_time = (time.time() - start) * 1000
            empty = len(result) == 0
            return {
                "success": not empty,
                "response_time": total_time,
                "empty": empty,
                "result": result,
            }
        except Exception as e:
            return {
                "success": False,
                "response_time": (time.time() - start) * 1000,
                "exception": e,
            }

    def delete(self, filter):
        start = time.time()
        try:
            result = self.client.delete(collection_name=self.collection_name, filter=filter)
            total_time = (time.time() - start) * 1000
            return {"success": True, "response_time": total_time, "result": result}
        except Exception as e:
            return {
                "success": False,
                "response_time": (time.time() - start) * 1000,
                "exception": e,
            }


# ----------------------------------
# Locust User wrapper
# ----------------------------------


class MilvusUser(User):
    """Locust User implementation for Milvus operations.

    This class wraps the MilvusV2Client implementation and translates
    client method results into Locust request events so that performance
    statistics are collected properly.

    Parameters
    ----------
    host : str
        Milvus server URI, e.g. ``"http://localhost:19530"``.
    collection_name : str
        The name of the collection to operate on.
    **client_kwargs
        Additional keyword arguments forwarded to the client.
    """

    abstract = True

    def __init__(
        self,
        environment,
        uri: str = "http://localhost:19530",
        token: str = "root:Milvus",
        collection_name: str = "test_collection",
        db_name: str = "default",
        timeout: int = 60,
        schema: CollectionSchema | None = None,
        index_params: IndexParams | None = None,
        **client_kwargs,
    ):
        super().__init__(environment)

        if uri is None:
            raise ValueError("'uri' must be provided for MilvusUser")
        if collection_name is None:
            raise ValueError("'collection_name' must be provided for MilvusUser")

        self.client_type = "milvus"
        self.client = MilvusV2Client(
            uri=uri,
            token=token,
            collection_name=collection_name,
            db_name=db_name,
            timeout=timeout,
        )
        if schema is not None:
            self.client.create_collection(schema=schema, index_params=index_params)

    @staticmethod
    def _fire_event(request_type: str, name: str, result: dict[str, Any]):
        """Emit a Locust request event from a Milvus client result dict."""
        response_time = int(result.get("response_time", 0))
        events.request.fire(
            request_type=f"{request_type}",
            name=name,
            response_time=response_time,
            response_length=0,
            exception=result.get("exception"),
        )

    @staticmethod
    def _fire_recall_event(request_type: str, name: str, result: dict[str, Any]):
        """Emit a Locust request event for recall metric using recall value instead of response time."""
        recall_value = result.get("recall", 0.0)
        # Use recall value as response_time for metric display (scaled by 100 for better visualization) percentage
        response_time_as_recall = int(recall_value * 100)
        events.request.fire(
            request_type=f"{request_type}",
            name=name,
            response_time=response_time_as_recall,
            response_length=result.get("retrieved_count", 0),
            exception=result.get("exception"),
        )

    def insert(self, data):
        result = self.client.insert(data)
        self._fire_event(self.client_type, "insert", result)
        return result

    def upsert(self, data):
        result = self.client.upsert(data)
        self._fire_event(self.client_type, "upsert", result)
        return result

    def search(
        self,
        data,
        anns_field,
        limit,
        filter="",
        search_params=None,
        output_fields=None,
        calculate_recall=False,
        ground_truth=None,
    ):
        result = self.client.search(
            data,
            anns_field,
            limit,
            filter=filter,
            search_params=search_params,
            output_fields=output_fields,
            calculate_recall=calculate_recall,
            ground_truth=ground_truth,
        )
        # Fire search event
        self._fire_event(self.client_type, "search", result)

        # Fire recall event if recall was calculated
        if calculate_recall and "recall" in result:
            self._fire_recall_event(self.client_type, "recall", result)

        return result

    def hybrid_search(self, reqs, ranker, limit, output_fields=None):
        result = self.client.hybrid_search(reqs, ranker, limit, output_fields)
        self._fire_event(self.client_type, "hybrid_search", result)
        return result

    def query(self, filter, output_fields=None):
        result = self.client.query(
            filter=filter,
            output_fields=output_fields,
        )
        self._fire_event(self.client_type, "query", result)
        return result

    def delete(self, filter):
        result = self.client.delete(filter)
        self._fire_event(self.client_type, "delete", result)
        return result

    def on_stop(self):
        self.client.close()
